
#include <stdlib.h>
#include <dos.h>
#include <stdio.h>

#include "cgdk.h"
#include "car.h"
#include "pcx.h"
#include "svga.h"


struct TPCXHeader {
    unsigned char       Manufacturer;
    unsigned char       Version;
    unsigned char       Encoding;
    unsigned char       BitsPerPixel;
    unsigned short      XMin;
    unsigned short      YMin;
    unsigned short      XMax;
    unsigned short      YMax;
    unsigned short      HDpi;
    unsigned short      VDpi;
    unsigned char       ColorMap[16][3];
    unsigned char       Reserved;
    unsigned char       NPlanes;
    unsigned short      BytesPerLine;
    unsigned short      PaletteInfo;
    unsigned short      HScreenSize;
    unsigned short      VScreenSize;
    unsigned char       pad[54];
};



static void *pcxBuf;
static long  pcxBufPos;
static long pcxBufLen;



//
//  Function: decodePCXLine
//    Description:
//      A module local function that decodes a lines-worth of data
//        from a PCX file.
//

static int decodePCXLine (unsigned char *lineBuffer, unsigned totalBytes)
{
  unsigned position;
  int count;
  unsigned char color;
  int rawData;
  int i;

  position = 0;

  while (position < totalBytes)
  {
    rawData = ((unsigned char *) pcxBuf) [pcxBufPos];
    pcxBufPos++;
    if (pcxBufPos > pcxBufLen)
    {
      return -1;
    }
    if ((rawData & 0xC0) == 0xC0)
    {
      count = rawData & 0x3F;
      color = ((unsigned char *) pcxBuf) [pcxBufPos];
      pcxBufPos++;
      if (pcxBufPos > pcxBufLen)
      {
        return -1;
      }
    }
    else
    {
      count = 1;
      color = rawData;
    }
    for (i = 0; i < count; i++)
    {
      lineBuffer[position++] = color;
    }
  }

  return 0;
}

//
//      Function: encodeRun
//    Description:
//      Encodes a run length and a data value and writes them to
//      a PCX file.  This routine has been adapted from the ZSoft
//        PCX documentation.  The routine returns a non-zero value
//        in the event of an error (typically disk full).
//

static int encodeRun (int val, int length, FILE *file)
{
  if (length)
  {
    if ((length == 1) && (0xC0 != (0xC0 & val)))
    {
      // run of 1 with color index < 0xC0 so no encoding //
      if (EOF == fputc (val, file))
      {
        return 1;
      }
    }
    else
    {
      if (EOF == fputc (0xC0 | length, file))
      {
        return 1;
      }
      if (EOF == fputc (val, file))
      {
        return 1;
      }
    }
  }
  return 0;
}

//
//      Function: encodePCXLine
//    Description:
//      Takes a line of byte data and run-length encodes it in PCX
//        format.  This routine has been adapted from the ZSoft PCX
//        documentation.  The routine returns a non-zero value in
//      the event of an error (typically disk full).
//

static int encodePCXLine (unsigned char *lineBuffer, unsigned Length, FILE *file)
{
  int oldVal;
  int newVal;
  int count;
  unsigned lengthCounter;

  // start us off //
  oldVal = *lineBuffer++;
  lengthCounter = 1;
  count = 1;

  while (lengthCounter < Length)
  {
    newVal = *lineBuffer++;
    if (newVal == oldVal)
    {
      // we've found a run //
      count++;
      if (count == 63)
      {
        // this run is at least 63 long, so save off what we //
        // have so far //
        if (encodeRun(oldVal, count, file))
        {
          return 1; // error //
        }
        count = 0;
      }
    }
    else
    {
      // no run //
      if (count)
      {
        // save off the run we were working on //
        if (encodeRun(oldVal, count, file))
        {
          return 1;
        }
      }
      oldVal = newVal;
      count = 1;
    }
    lengthCounter++;
  }

  // save off last run //
  if (count)
  {
    if (encodeRun(oldVal, count, file))
    {
      return 1;
    }
  }

  return 0;
}


//
//      Function: PCX_save
//    Description:
//      Take a linear bitmap and a palette as input and saves the
//      linear bitmap as a PCX file with the specified name.
//      The bitmap will be saved as a version 5 format PCX file with
//      a palette.
//


int PCX_save (char *fileName,
              char *videoBuf, int sizeX, int sizeY,
              TRGBColor pal [256])
{
  TPCXHeader pcxHead;
  FILE * file;
  int colorIndex;
  int i;
  int result;
  unsigned lineCounter;
  unsigned char *data;
  int palData;

  // fill in PCX header //
  pcxHead.Manufacturer      = 10;
  pcxHead.Version           = 5;
  pcxHead.Encoding          = 1;
  pcxHead.BitsPerPixel      = 8;
  pcxHead.XMin                      = 0;
  pcxHead.YMin                      = 0;
      pcxHead.XMax                  = sizeX - 1;
  pcxHead.YMax                      = sizeY - 1;
  pcxHead.HDpi          = sizeX;
  pcxHead.VDpi                      = sizeY;
  pcxHead.Reserved          = 0;
  pcxHead.NPlanes           = 1;
  pcxHead.BytesPerLine      = (sizeX + 1) & 0xFFFE; // make even //
  pcxHead.PaletteInfo       = 1;
  pcxHead.HScreenSize       = sizeX;
  pcxHead.VScreenSize       = sizeY;
  for (colorIndex = 0; colorIndex < 16; colorIndex++) {
          pcxHead.ColorMap[colorIndex][0] = pal[colorIndex].r;
          pcxHead.ColorMap[colorIndex][1] = pal[colorIndex].g;
          pcxHead.ColorMap[colorIndex][2] = pal[colorIndex].b;
  }
  for (i = 0; i < sizeof(pcxHead.pad); i++) {
      pcxHead.pad[i] = 0;
  }

  // open file //
  file = fopen(fileName, "wb");
  if (file == NULL) {
      return 1; // error //
  }

  // write header to file //
  result = fwrite(&pcxHead, sizeof(TPCXHeader), 1, file);
  if (result != 1) {
      // write error //
      fclose(file);
      return 1;
  }

  // encode data //
  data = (unsigned char  *) videoBuf;
  for (lineCounter = 0; lineCounter < sizeY; lineCounter++) {
      if (encodePCXLine(data, sizeX, file)) {
              fclose(file);
              return 1; // file write error //
              }
      // handle odd width since BytesPerLine must be even //
      if (sizeX & 1) {
              if (fputc(0, file) != 0) {
              fclose(file);
              return 1; // file write error //
          }
      }
      data += sizeX;
  }

  // write palette //
  fputc(12, file);    // palette marker //
  for (colorIndex = 0; colorIndex < 256; colorIndex++) {
      palData = pal[colorIndex].r;
      result = fputc(palData, file);
      if (result != palData) {
              fclose(file);
              return 1; // write error //
      }
      palData = pal[colorIndex].g;
      result = fputc(palData, file);
      if (result != palData) {
              fclose(file);
              return 1; // write error //
      }
      palData = pal[colorIndex].b;
      result = fputc(palData, file);
      if (result != palData) {
              fclose(file);
              return 1; // write error //
      }
  }

  // close and exit //
  fclose(file);
  return 0;
}


int PCX_getInfo (TCAR *carHandle, const char *fileName, int *xSize, int *ySize, int *bytesPerPixel)
{
  TPCXHeader pcxHead;

  if (CAR_seek (carHandle, fileName, 0) == -1)
  {
    return 1;
  }
  else
  {
    if (CAR_read (carHandle, &pcxHead, sizeof (pcxHead)) != sizeof (pcxHead))
      return 2;

    switch (pcxHead.BitsPerPixel)
    {
      case 8:
      {
        *xSize = pcxHead.XMax - pcxHead.XMin + 1;
        *ySize = pcxHead.YMax - pcxHead.YMin + 1;
        *bytesPerPixel = pcxHead.BytesPerLine / *xSize;
        break;
      }
      default:
      {
        return 3;
      }
    }

    return 0;
  }
}





int PCX_load (TCAR *carHandle, const char *fileName, void far *videoBuf,
                             int linearBufSizeX, int linearBufSizeY,
                             TRGBColor pal [256])
{
    unsigned char *lineBuffer;
    unsigned char far *linearDataPointer;
    long fileLen;


    TPCXHeader pcxHead;
    unsigned lineCounter;
    unsigned widthCounter;
    int result;
    unsigned totalBytes;
    int colorIndex;
    short xSize, ySize;
    int i;

    if (videoBuf == NULL)
      return 1;

    // read in header //
    if ((fileLen = CAR_seek (carHandle, fileName, 0)) == -1)
    {
      return 2;
    }

    result = CAR_read (carHandle, &pcxHead, sizeof(pcxHead));
    if (result != sizeof(pcxHead)) {
      return 3;
    }

    pcxBufLen = fileLen - sizeof (pcxHead);


    // compute size of image and scanline size //
    xSize = pcxHead.XMax - pcxHead.XMin + 1;
    ySize = pcxHead.YMax - pcxHead.YMin + 1;
    totalBytes = pcxHead.NPlanes * pcxHead.BytesPerLine;

    // allocate scanline buffer //
    lineBuffer = (unsigned char  *) _alloc(totalBytes);
    if (lineBuffer == NULL) {
        return 4;
    }

    pcxBuf = _alloc ((unsigned)pcxBufLen);

    if (pcxBuf == NULL)
    {
      _free (lineBuffer);
      lineBuffer = NULL;
      return 5;
    }

    pcxBufPos = 0;

    if (CAR_read (carHandle, pcxBuf, pcxBufLen) != pcxBufLen)
    {
      _free (lineBuffer);
      lineBuffer = NULL;
      _free (pcxBuf);
      pcxBuf = NULL;
      return 6;
    }


    // decode data //
    linearDataPointer = (unsigned char far *) videoBuf;

    for (lineCounter = 0; (lineCounter < ySize) &&
                          (lineCounter < linearBufSizeY);
                          lineCounter++)
    {
        result = decodePCXLine(lineBuffer, totalBytes);
        if (result != 0) {
            _free (pcxBuf);
            pcxBuf = NULL;
            _free (lineBuffer);
            lineBuffer = NULL;
            return 7;
        }
        for (widthCounter = 0; (widthCounter < xSize) &&
                               (widthCounter < linearBufSizeX);
                               widthCounter++)
        {
          linearDataPointer [widthCounter] = lineBuffer[pcxHead.XMin + widthCounter];
        }
        linearDataPointer += linearBufSizeX;
    }

    _free(lineBuffer);
    lineBuffer = NULL;

    // if the caller doesn't want palette information, just return //
    if (pal == NULL)
    {
      return 0;
    }

    // get the palette //
    if (pcxHead.Version != 5) {
            // if the PCX file doesn't support the 256-color palette //
        // use the 16-color palette instead //
        for (colorIndex = 0; colorIndex < 16; colorIndex++) {
            pal [colorIndex].r = pcxHead.ColorMap[colorIndex][0];
            pal [colorIndex].g = pcxHead.ColorMap[colorIndex][1];
            pal [colorIndex].b = pcxHead.ColorMap[colorIndex][2];
            pal [colorIndex].a = 0;
        }
        for (colorIndex = 16; colorIndex < 256; colorIndex++) {
            pal [colorIndex].r = 0;
            pal [colorIndex].g = 0;
            pal [colorIndex].g = 0;
            pal [colorIndex].a = 0;
        }
    }
    else {
        result = ((char *) pcxBuf) [pcxBufLen-(256*3+1)];
        if (result != 12) {
                // this doesn't have a 256-color palette, //
            // so use the 16-color one //
                for (colorIndex = 0; colorIndex < 16; colorIndex++)
                {
                  // copy entries and divide by four since PCX colors //
                  // can range from 0 to 255 and VGA colors can only //
                  // go as high as 63 //
                  pal [colorIndex].r = pcxHead.ColorMap[colorIndex][0];
                  pal [colorIndex].g = pcxHead.ColorMap[colorIndex][1];
                  pal [colorIndex].b = pcxHead.ColorMap[colorIndex][2];
                }
                for (colorIndex = 16; colorIndex < 256; colorIndex++) {
                pal [colorIndex].r = 0;
                pal [colorIndex].g = 0;
                pal [colorIndex].b = 0;
                pal [colorIndex].a = 0;
            }
        }
        else {
            i = pcxBufLen - 256*3;
            for (colorIndex = 0; colorIndex < 256; colorIndex++) {
                pal[colorIndex].r = ((char *)pcxBuf) [i];
                i++;
                pal[colorIndex].g = ((char *)pcxBuf) [i];
                i++;
                pal[colorIndex].b = ((char *)pcxBuf) [i];
                i++;
                pal[colorIndex].a = 0;
            }
        }
    }

    _free (pcxBuf);
    pcxBuf = NULL;
    return 0;
}


TRGBColor *PCX_pal;

int PCX_view (TCAR *carHandle, const char *fileName, int x, int y)
{
  int PCX_sizeX, PCX_sizeY;
  int sizeX, sizeY;
  int bytesPerPixel;
  int exitCode;
  void far *pcxBuf;

  if (PCX_getInfo (carHandle, fileName, &PCX_sizeX, &PCX_sizeY, &bytesPerPixel) == 0)
  {
    // do some adjust to center the image on the screen
    if (x == PCX_VIEW_CENTER)
    {
      x = (SVGA_maxX - PCX_sizeX)/2;
    }

    if (y == PCX_VIEW_CENTER)
    {
      y = (SVGA_maxY - PCX_sizeY)/2;
    }
    if (x < 0)
      x = 0;
    if (y < 0)
      y = 0;

    if (SVGA_linearBuf == NULL)
    {
      sizeX = PCX_sizeX;
      sizeY = PCX_sizeY;
      pcxBuf = (void far *) _alloc ((unsigned)(SVGA_virtualX*SVGA_maxY*bytesPerPixel));
      if (pcxBuf == NULL)
      {
        return -3;
      }
    }
    else
    {
      sizeX = SVGA_virtualX;
      sizeY = SVGA_maxY-y;
      pcxBuf = (char far *) SVGA_linearBuf+x+y*SVGA_virtualX;
    }

    if (PCX_pal == NULL)
    {
      PCX_pal = (TRGBColor *) _alloc (256*sizeof (TRGBColor));
      if (PCX_pal == NULL)
      {
        _free ((void *)pcxBuf);
        return -4;
      }
    }

    if (PCX_load (carHandle, fileName, pcxBuf,
                  sizeX, sizeY, PCX_pal) == 0)
    {
      if (SVGA_linearBuf == NULL)
      {
        SVGA_putWin ((void *)pcxBuf, x, y, sizeX, sizeY, sizeX);
      }
      exitCode = 0;
    }
    else
    {
      exitCode = -2;
    }
    if (SVGA_linearBuf == NULL)
    {
      _free ((void *) pcxBuf);
    }
  }
  else
  {
    exitCode = -1;
  }
  return exitCode;
}
